#include <stdio.h>

#include "world.h"
#include "creature.h"

#include "action.h"

#define CREATURE_IFF_UNKNOWN 0
#define CREATURE_IFF_FRIENDLY 1
#define CREATURE_IFF_HOSTILE 2

static int get_creature_iff(world *wp, int creature_id);

static int
get_creature_iff(world *wp, int creature_id)
{
  if (wp == NULL)
    return CREATURE_IFF_UNKNOWN;

  if ((creature_id < 0) || (creature_id >= wp->cr_size))
    return CREATURE_IFF_UNKNOWN;
  if (wp->cr[creature_id] == NULL)
    return CREATURE_IFF_UNKNOWN;

  if (creature_id == wp->player_id)
    return CREATURE_IFF_FRIENDLY;

  return CREATURE_IFF_HOSTILE;
}

/* return 1 if the action is legal, 0 otherwise */
int
check_walk(world *wp, int creature_id, int dx, int dy)
{
  int target_x;
  int target_y;
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "check_walk: wp is NULL\n");
    return 0;
  }

  if ((creature_id < 0) || (creature_id >= wp->cr_size))
    return 0;
  if (wp->cr[creature_id] == NULL)
    return 0;
  who = wp->cr[creature_id];

  if (who->hp <= 0)
    return 0;

  if ((dx == 0) && (dy == 0))
    return 1;

  if ((dx < -1) || (dx > 1))
    return 0;
  if ((dy < -1) || (dy > 1))
    return 0;

  target_x = who->x + dx;
  target_y = who->y + dy;

  if ((target_x < 0) || (target_x >= wp->size_x))
    return 0;
  if ((target_y < 0) || (target_y >= wp->size_y))
    return 0;

  if (wp->grid[target_x][target_y])
    return 0;

  /* assumes that:
   * * no creature is in a wall
   * * there is no friendly non-player character
   */
  return 1;
}

int
exec_walk(world *wp, int creature_id)
{
  int i;
  int target_x;
  int target_y;
  int blocked;
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "exec_walk: wp is NULL\n");
    return 1;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "exec_walk: strange creature_id (%d)\n", creature_id);
    return 1;
  }
  who = wp->cr[creature_id];

  if (!check_walk(wp, who->id, who->dx, who->dy))
  {
    fprintf(stderr, "exec_walk: creature %d tries something illegal "
            "(%d, %d)\n",
            who->id, who->dx, who->dy);
    return 1;
  }

  if ((who->dx == 0) && (who->dy == 0))
    return 0;

  target_x = who->x + who->dx;
  target_y = who->y + who->dy;

  blocked = 0;
  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == who->id)
      continue;
    if (wp->cr[i] == NULL)
      continue;
    if (wp->cr[i]->hp <= 0)
      continue;
    if (wp->cr[i]->x != target_x)
      continue;
    if (wp->cr[i]->y != target_y)
      continue;

    blocked = 1;

    if (get_creature_iff(wp, i) == get_creature_iff(wp, who->id))
      continue;

    if (i == wp->player_id)
    {
      /* player is invulnerable */
      (wp->player_damage)++;
      while (wp->player_damage >= 100)
        wp->player_damage -= 10;
    }
    else
    {
      wp->cr[i]->hp = 0;
    }
  }

  if (wp->grid[target_x][target_y])
    blocked = 1;

  if (!blocked)
  {
    who->x = target_x;
    who->y = target_y;
  }

  return 0;
}
