#include <stdio.h>
/* atexit */
#include <stdlib.h>

#include <curses.h>

#include "world.h"
#include "color-pair.h"

#include "init-curses.h"

static void curses_cleanup(void);

static
void curses_cleanup(void)
{
  clear();
  refresh();
  nocbreak();
  echo();
  nl();
  endwin();
}

int
init_curses(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "init_curses: wp is NULL\n");
    return 1;
  }

  wp->scr = initscr();
  if (wp->scr == NULL)
  {
    fprintf(stderr, "init_curses: initscr failed\n");
    return 1;
  }

  if (atexit(&curses_cleanup) != 0)
  {
    fprintf(stderr, "init_curses: atexit failed\n");
    curses_cleanup();
    return 1;
  }

  start_color();
  cbreak();
  noecho();
  nonl();
  keypad(wp->scr, TRUE);

  init_pair(CPAIR_RED, COLOR_RED, COLOR_BLACK);
  init_pair(CPAIR_GREEN, COLOR_GREEN, COLOR_BLACK);
  init_pair(CPAIR_YELLOW, COLOR_YELLOW, COLOR_BLACK);
  init_pair(CPAIR_BLUE, COLOR_BLUE, COLOR_BLACK);
  init_pair(CPAIR_MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
  init_pair(CPAIR_CYAN, COLOR_CYAN, COLOR_BLACK);
  init_pair(CPAIR_WHITE, COLOR_WHITE, COLOR_BLACK);
  init_pair(CPAIR_BLACK, COLOR_BLACK, COLOR_BLACK);

  getmaxyx(wp->scr, wp->scr_size_y, wp->scr_size_x);

  clear();
  refresh();

  return 0;
}
