/* FILE */
#include <stdio.h>
/* errno */
#include <errno.h>
/* strerror */
#include <string.h>
/* isdigit, isalpha */
#include <ctype.h>

#include "world.h"

#include "load.h"

int
load_map(world *wp, const char *map_file)
{
  int x;
  int y;
  int c;
  int num_monster;
  FILE *fp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "load_map: wp is NULL\n");
    return 1;
  }
  if (map_file == NULL)
  {
    fprintf(stderr, "load_map: map_file is NULL\n");
    return 1;
  }

  errno = 0;
  fp = fopen(map_file, "r");
  if (fp == NULL)
  {
    fprintf(stderr, "load_map: can't open %s", map_file);
    if (errno != 0)
      fprintf(stderr, " (%s)", strerror(errno));
    fprintf(stderr, "\n");
    return 1;
  }

  wp->player_id = 0;
  num_monster = 0;

  x = 0;
  y = 0;

  while(y < wp->size_y)
  {
    c = getc(fp);
    if (c == EOF)
      break;

    if (c == '\n')
    {
      y++;
      x = 0;
      continue;
    }

    if (x >= wp->size_x)
    {
      continue;
    }

    if ((c == '@') || (isdigit(c)))
    {
      wp->cr[wp->player_id]->x = x;
      wp->cr[wp->player_id]->y = y;
      wp->cr[wp->player_id]->hp = 1;

      wp->grid[x][y] = 0;
    }
    else if (isalpha(c))
    {
      /* +1 is for the player character */
      if (num_monster + 1 < wp->cr_size)
      {
        /* assumes that wp->player_id is 0 */
        wp->cr[num_monster + 1]->x = x;
        wp->cr[num_monster + 1]->y = y;
        wp->cr[num_monster + 1]->hp = 1;
        (num_monster)++;
      }

      wp->grid[x][y] = 0;
    }
    else if (c == '#')
    { 
      wp->grid[x][y] = 1;
    }
    else
    {      
      wp->grid[x][y] = 0;
    }

    x++;
  }

  fclose(fp);
  fp = NULL;

  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size)
      || (wp->cr[wp->player_id] == NULL)
      || (wp->cr[wp->player_id]->hp <= 0))
  {
    fprintf(stderr, "load_map: player character not found\n");
    return 1;
  }
  if ((wp->cr[wp->player_id]->x < 0)
      || (wp->cr[wp->player_id]->x >= wp->size_x)
      || (wp->cr[wp->player_id]->y < 0)
      || (wp->cr[wp->player_id]->y >= wp->size_y))
  {
    fprintf(stderr, "load_map: player character (%d, %d)is out of map "
            "(%d x %d)\n",
            wp->cr[wp->player_id]->x, wp->cr[wp->player_id]->y,
            wp->size_x, wp->size_y);
    return 1;
  }

  return 0;
}
