#include <stdio.h>
#include <curses.h>

#include "world.h"
#include "color-pair.h"
#include "creature.h"

#include "loop-output.h"

static int loop_output_grid(world *wp);
static int loop_output_creature(world *wp);

static int
loop_output_grid(world *wp)
{
  int x;
  int y;
  int c;
  int attr;

  if (wp == NULL)
    return 1;

  attr = COLOR_PAIR(CPAIR_WHITE);

  attron(attr);
  for (x = 0; x < wp->size_x; x++)
  {
    for (y = 0; y < wp->size_y; y++)
    {
      if (wp->grid[x][y])
        c = '#';
      else
        c = '.';
      mvaddch(y, x, c);
    }
  }
  attroff(attr);

  return 0;
}

static int
loop_output_creature(world *wp)
{
  int i;
  int c;
  int attr;
  creature *who = NULL;

  if (wp == NULL)
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    who = wp->cr[i];
    if (who == NULL)
      continue;
    if (who->hp <= 0)
      continue;

    if (i == wp->player_id)
    {
      /* assumes the ASCII code */
      c = '0' + (wp->player_damage % 10);
      attr = COLOR_PAIR(CPAIR_WHITE) | A_BOLD;
    }
    else
    {
      /* assumes the ASCII code */
      c = 'a' + (i % 26);
      attr = COLOR_PAIR(CPAIR_WHITE);
    }

    attron(attr);
    mvaddch(who->y, who->x, c);
    attroff(attr);
  }

  return 0;
}

int
loop_output(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "loop_output: wp is NULL\n");
    return 1;
  }

  erase();

  loop_output_grid(wp);
  loop_output_creature(wp);

  move(wp->scr_size_y - 1, 0);
  refresh();

  return 0;
}
