#include <stdio.h>
/* srand */
#include <stdlib.h>
/* time */
#include <time.h>
/* strcmp */
#include <string.h>

#include "world.h"
#include "init-curses.h"
#include "loop.h"
#include "build.h"

#define DEFAULT_MAP_FILE "map.txt"

static int do_version(void);
static int do_help(void);

static int
do_version(void)
{
  printf("monsters surround player tech demo\n");
  printf("Tue, 05 Feb 2013 release\n");
  printf("Copyright (C) 2013 Oohara Yuuma <oohara@libra.interq.or.jp>\n");
  printf("This program is free software;\n");
  printf("see the file LICENSE in the source code for details.\n");
  printf("This program comes with NO WARRANTY, to the extent "
         "permitted by law.\n");

  return 0;
}

static int
do_help(void)
{
  printf("Usage: surround [options]\n");
  printf("see how the monsters block the other side of a corridor\n");
  printf("Options:\n");
  printf("  --help              print this message\n");
  printf("  --version           print version information\n");
  printf("  --map-file _file_   read map from this file\n");
  printf("    (default: %s)\n", DEFAULT_MAP_FILE);
  printf("  --no-surround       always follow the shortest path\n");
  printf("  --no-hesitate       attack player as soon as possible\n");
  printf("\n");
  printf("Report bugs to <oohara@libra.interq.or.jp>.\n");

  return 0;
}

int
main(int argc, char *argv[])
{
  int i;
  int flag_help;
  int flag_version;
  int flag_no_surround;
  int flag_no_hesitate;
  world *wp = NULL;
  const char *map_file = NULL;

  flag_help = 0;
  flag_version = 0;
  flag_no_surround = 0;
  flag_no_hesitate = 0;
  map_file = DEFAULT_MAP_FILE;
  for (i = 1; i < argc; i++)
  {
    if (!strcmp(argv[i], "--help"))
    {
      flag_help = 1;
    }
    else if (!strcmp(argv[i], "--version"))
    {
      flag_version = 1;
    }
    else if (!strcmp(argv[i], "--map-file"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for --map_file\n");
        return 1;
      }
      map_file = argv[i + 1];
      i++;
    }
    else if (!strcmp(argv[i], "--no-surround"))
    {
      flag_no_surround = 1;
    }
    else if (!strcmp(argv[i], "--no-hesitate"))
    {
      flag_no_hesitate = 1;
    }
    else
    {
      fprintf(stderr, "main: unknown option: %s\n", argv[i]);
      fprintf(stderr, "Type %s --help for the list of options.\n", argv[0]);
      return 1;
    }
  }

  if (flag_version)
    do_version();
  if (flag_help)
    do_help();
  if ((flag_help) || (flag_version))
    return 0;

  srand((unsigned int) time(NULL));

  wp = world_new();
  if (wp == NULL)
  {
    fprintf(stderr, "main: world_new failed\n");
    return 1;
  }

  if (flag_no_surround)
    wp->no_surround = 1;
  if (flag_no_hesitate)
    wp->no_hesitate = 1;

  if (build_game(wp, map_file) != 0)
  {
    fprintf(stderr, "main: build_game failed\n");
    world_delete(wp);
    wp = NULL;
    return 1;
  }

  if (init_curses(wp) != 0)
  {
    fprintf(stderr, "main: init_curses failed\n");
    world_delete(wp);
    wp = NULL;
    return 1;
  }

  main_loop(wp);

  world_delete(wp);
  wp = NULL;

  return 0;
}
