#include <stdio.h>
/* rand */
#include <stdlib.h>

#include "world.h"
#include "action.h"
#include "creature.h"
#include "dijkstra.h"
#include "map-blocked.h"

#include "monster.h"

static int cardinal_direction(int dx, int dy);
static int monster_decide_action_1(world *wp);
static int monster_decide_action_2(world *wp);
static int monster_decide_action_3(world *wp);

/* return 1 (true) or 0 (false) */
static int
cardinal_direction(int dx, int dy)
{
  if (dx == 0)
    return 1;
  if (dy == 0)
    return 1;
  if (dx == dy)
    return 1;
  if (dx == -dy)
    return 1;

  return 0;
}

/* prepare Dijkstra maps */
static int
monster_decide_action_1(world *wp)
{
  int i;
  int j;
  int dx;
  int dy;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  if (wp->cr[wp->player_id] == NULL)
  {
    fprintf(stderr, "monster_decide_action_1: player is NULL\n");
    return 1;
  }
  player = wp->cr[wp->player_id];

  dijkstra_map_reset(wp->dijkp[4]);
  dijkstra_map_init(wp->dijkp[4], player->x + 0, player->y + 0);
  dijkstra_map_refresh(wp->dijkp[4], wp->map_blocked, -1);

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (wp->cr[i] == NULL)
      continue;
    if (wp->cr[i]->hp <= 0)
      continue;

    wp->map_blocked[wp->cr[i]->x][wp->cr[i]->y] |= MAP_BLOCKED_MONSTER;
  }

  for (i = 0; i < 9; i++)
  {
    if (i == 4)
      continue;

    for (j = 0; j < 9; j++)
    {
      if (j == i)
        continue;

      dx = -1 + (j % 3);
      dy = 1 - (j / 3);

      wp->map_blocked[player->x + dx][player->y + dy]
        |= MAP_BLOCKED_NEAR_PLAYER;
    }

    dx = -1 + (i % 3);
    dy = 1 - (i / 3);

    dijkstra_map_reset(wp->dijkp[i]);
    dijkstra_map_init(wp->dijkp[i], player->x + dx, player->y + dy);
    dijkstra_map_refresh(wp->dijkp[i], wp->map_blocked, -1);

    for (j = 0; j < 9; j++)
    {
      if (j == i)
        continue;

      dx = -1 + (j % 3);
      dy = 1 - (j / 3);

      wp->map_blocked[player->x + dx][player->y + dy]
        &= ~MAP_BLOCKED_NEAR_PLAYER;
    }
  }

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (wp->cr[i] == NULL)
      continue;
    if (wp->cr[i]->hp <= 0)
      continue;

    wp->map_blocked[wp->cr[i]->x][wp->cr[i]->y] &= ~MAP_BLOCKED_MONSTER;
  }

  return 0;
}

/* when engaging, don't be in the way of other monsters */
static int
monster_decide_action_2(world *wp)
{
  int i;
  int j;
  int x;
  int y;
  int dx;
  int dy;
  int x0;
  int y0;
  int m;
  int n;
  int num_dir_blocked;
  int monster_id;
  creature *player = NULL;
  int dir_blocked[9];

  if (wp == NULL)
    return 1;

  if (wp->cr[wp->player_id] == NULL)
  {
    fprintf(stderr, "monster_decide_action_2: player is NULL\n");
    return 1;
  }
  player = wp->cr[wp->player_id];

  for (i = 0; i < 9; i++)
  {
    if (i == 4)
    {
      dir_blocked[i] = 1;
      continue;
    }

    dx = -1 + (i % 3);
    dy = 1 - (i / 3);

    x = player->x + dx;
    y = player->y + dy;

    if ((x < 0) || (x >= wp->size_x)
        || (y < 0) || (y >= wp->size_y))
    {
      dir_blocked[i] = 1;
    }
    else
    {
      dir_blocked[i] = wp->grid[x][y];
    }
  }

  wp->num_engage = 0;
  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (wp->cr[i] == NULL)
      continue;
    if (wp->cr[i]->hp <= 0)
      continue;

    dx = wp->cr[i]->x - player->x;
    dy = wp->cr[i]->y - player->y;

    if ((dx < -2) || (dx > 2)
        || (dy < -2) || (dy > 2))
      continue;

    if ((dx >= -1) && (dx <= 1)
        && (dy >= -1) && (dy <= 1))
    {
      dir_blocked[(1 + dx) + 3 - dy * 3] = 1;
      continue;
    }

    dx += wp->cr[i]->dx;
    dy += wp->cr[i]->dy;

    if ((dx < -1) || (dx > 1)
        || (dy < -1) || (dy > 1))
      continue;

    wp->engage_id[wp->num_engage] = i;
    wp->engage_num_choice[wp->num_engage] = 0;
    (wp->num_engage)++;

    wp->cr[i]->dx = 0;
    wp->cr[i]->dy = 0;
  }

  num_dir_blocked = 0;
  for (i = 0; i < 9; i++)
  {
    if (dir_blocked[i])
      num_dir_blocked++;
  }

  if (num_dir_blocked >= 9)
    return 0;

  /* the order does matter here */
  for (i = wp->num_engage - 1; i >= 0; i--)
  {
    x0 = wp->cr[wp->engage_id[i]]->x;
    y0 = wp->cr[wp->engage_id[i]]->y;

    for (j = 0; j < 9; j++)
    {
      if (j == 4)
        continue;

      if (dir_blocked[j])
        continue;

      dx = -1 + (j % 3);
      dy = 1 - (j / 3);

      x = player->x + dx;
      y = player->y + dy;

      if ((x - x0 < -1) || (x - x0 > 1)
          || (y - y0 < -1) || (y - y0 > 1))
        continue;

      (wp->engage_num_choice[i])++;
    }

    if (wp->engage_num_choice[i] <= 0)
    {
      wp->engage_id[i] = wp->engage_id[wp->num_engage - 1];
      wp->engage_num_choice[i] = wp->engage_num_choice[wp->num_engage - 1];
      (wp->num_engage)--;
    }
  }

  for (i = 0; i < wp->num_engage - 1; i++)
  {
    n = i + (rand() % (wp->num_engage - i));
    if (n == i)
      continue;

    m = wp->engage_id[i];
    wp->engage_id[i] = wp->engage_id[n];
    wp->engage_id[n] = m;

    m = wp->engage_num_choice[i];
    wp->engage_num_choice[i] = wp->engage_num_choice[n];
    wp->engage_num_choice[n] = m;
  }

  for (i = 0; i < wp->num_engage; i++)
  {
    if (num_dir_blocked >= 9)
      break;

    n = -1;
    monster_id = -1;
    for (j = 0; j < wp->num_engage; j++)
    {
      if (wp->engage_num_choice[j] <= 0)
        continue;

      if ((monster_id < 0)
          || (wp->engage_num_choice[j] < wp->engage_num_choice[n]))
      {
        n = j;
        monster_id = wp->engage_id[j];
      }
    }

    if (monster_id < 0)
      break;

    if ((monster_id >= wp->cr_size) || (monster_id == wp->player_id))
    {
      fprintf(stderr, "monster_decide_action_2: strange monster_id (%d\n)",
              monster_id);
      break;
    }

    wp->engage_num_choice[n] = 0;

    x0 = wp->cr[monster_id]->x;
    y0 = wp->cr[monster_id]->y;

    m = -1;
    n = 0;
    for (j = 0; j < 9; j++)
    {
      if (j == 4)
        continue;

      if (dir_blocked[j])
        continue;

      dx = -1 + (j % 3);
      dy = 1 - (j / 3);

      x = player->x + dx;
      y = player->y + dy;

      if ((x - x0 < -1) || (x - x0 > 1)
          || (y - y0 < -1) || (y - y0 > 1))
        continue;

      n++;
      if (rand() % n == 0)
      {
        wp->cr[monster_id]->dx = x - x0;
        wp->cr[monster_id]->dy = y - y0;

        m = j;
      }
    }

    if (m < 0)
    {
      fprintf(stderr, "monster_decide_action_2: m is negative; "
              "should not happen\n");
      break;
    }

    dir_blocked[m] = 1;
    num_dir_blocked++;

    x = player->x + (-1 + (m % 3));
    y = player->y + (1 - (m / 3));

    for (j = 0; j < wp->num_engage; j++)
    {
      if (wp->engage_num_choice[j] <= 0)
        continue;

      x0 = wp->cr[wp->engage_id[j]]->x;
      y0 = wp->cr[wp->engage_id[j]]->y;

      if ((x - x0 >= -1) && (x - x0 <= 1)
          && (y - y0 >= -1) && (y - y0 <= 1))
      {
        (wp->engage_num_choice[j])--;
      }
    }
  }

  return 0;
}

/* don't engage alone */
static int
monster_decide_action_3(world *wp)
{
  int i;
  int dx;
  int dy;
  int n;
  int engage;
  creature *player = NULL;
  int dir[9];

  if (wp == NULL)
    return 1;

  if (wp->cr[wp->player_id] == NULL)
  {
    fprintf(stderr, "monster_decide_action_3: player is NULL\n");
    return 1;
  }
  player = wp->cr[wp->player_id];

  engage = 0;
  for (i = 0; i < 9; i++)
    dir[i] = 0;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (wp->cr[i] == NULL)
      continue;
    if (wp->cr[i]->hp <= 0)
      continue;

    dx = (wp->cr[i]->x + wp->cr[i]->dx) - player->x;
    dy = (wp->cr[i]->y + wp->cr[i]->dy) - player->y;

    if ((dx < -1) || (dx > 1))
      continue;
    if ((dy < -1) || (dy > 1))
      continue;

    n = (dx + 1) + 3 - dy * 3;

    if (!(dir[n]))
    {
      dir[n] = 1;
      engage++;
    }
  }

  if (engage != 1)
    return 0;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (wp->cr[i] == NULL)
      continue;
    if (wp->cr[i]->hp <= 0)
      continue;

    dx = wp->cr[i]->x - player->x;
    dy = wp->cr[i]->y - player->y;

    if ((dx >= -1) && (dx <= 1)
        && (dy >= -1) && (dy <= 1))
      continue;

    dx += wp->cr[i]->dx;
    dy += wp->cr[i]->dy;

    if ((dx >= -1) && (dx <= 1)
        && (dy >= -1) && (dy <= 1))
    {
      wp->cr[i]->dx = 0;
      wp->cr[i]->dy = 0;
    }
  }

  return 0;
}

int
monster_decide_action(world *wp)
{
  int i;
  int j;
  int k;
  int x0;
  int y0;
  int dx1;
  int dy1;
  int dx2;
  int dy2;
  int m;
  int n;
  int length;
  int strictly_better;
  creature *player = NULL;
  dijkstra_map *dijkp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "monster_decide_action: wp is NULL\n");
    return 1;
  }

  if (wp->cr[wp->player_id] == NULL)
  {
    fprintf(stderr, "monster_decide_action: player is NULL\n");
    return 1;
  }
  player = wp->cr[wp->player_id];

  monster_decide_action_1(wp);

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (wp->cr[i] == NULL)
      continue;
    if (wp->cr[i]->hp <= 0)
      continue;

    x0 = wp->cr[i]->x;
    y0 = wp->cr[i]->y;

    if ((player->x - x0 >= -1) && (player->x - x0 <= 1)
        && (player->y - y0 >= -1) && (player->y - y0 <= 1))
    {
      wp->cr[i]->dx = player->x - x0;
      wp->cr[i]->dy = player->y - y0;
      continue;
    }

    for (j = 0; j < 9; j++)
    {
      m = wp->cr[i]->id;
      if (wp->cr[i]->id % 18 < 9)
      { 
        m += j;
      }
      else
      { 
        m -= j;
        m += 9;
      }
      m %= 9;

      if (wp->no_surround)
        m = 4;

      dx1 = -1 + (m % 3);
      dy1 = 1 - (m / 3);

      if ((x0 == player->x + dx1) && (y0 == player->y + dy1))
      {
        wp->cr[i]->dx = 0;
        wp->cr[i]->dy = 0;
        break;
      }

      if (wp->grid[player->x + dx1][player->y + dy1])
        continue;

      dijkp = wp->dijkp[m];

      length = dijkp->result[x0][y0];
      if (length >= 0)
      {
        wp->cr[i]->dx = 0;
        wp->cr[i]->dy = 0;
        n = 1;
      }
      else
      {
        n = 0;
      }

      for (k = 0; k < 9; k++)
      {
        if (k == 4)
          continue;

        dx2 = -1 + (k % 3);
        dy2 = 1 - (k / 3);

        if (wp->grid[x0 + dx2][y0 + dy2])
          continue;
        if (dijkp->result[x0 + dx2][y0 + dy2] < 0)
          continue;

        strictly_better = 0;
        if (length >= 0)
        {
          if (length < dijkp->result[x0 + dx2][y0 + dy2])
          {
            continue;
          }
          else if (length > dijkp->result[x0 + dx2][y0 + dy2])
          { 
            strictly_better = 1;
          }
          else if ((x0 - player->x < -3) || (x0 - player->x > 3)
                   || (y0 - player->y < -3) || (y0 - player->y > 3))
          {
            if ((!cardinal_direction(x0 + wp->cr[i]->dx - player->x,
                                     y0 + wp->cr[i]->dy - player->y))
                && (cardinal_direction(x0 + dx2 - player->x,
                                       y0 + dy2 - player->y)))
            {
              continue;
            }
            else if ((cardinal_direction(x0 + wp->cr[i]->dx - player->x,
                                         y0 + wp->cr[i]->dy - player->y))
                     && (!cardinal_direction(x0 + dx2 - player->x,
                                             y0 + dy2 - player->y)))
            {
              strictly_better = 1;
            }
          }
        }
        else
        {
          strictly_better = 1;
        }

        if (strictly_better)
        {
          n = 1;
        }
        else
        {
          n++;
        }

        length = dijkp->result[x0 + dx2][y0 + dy2];

        if (rand() % n == 0)
        {
          wp->cr[i]->dx = dx2;
          wp->cr[i]->dy = dy2;
        }
      }

      if (n > 0)
        break;

      if (wp->no_surround)
        break;
    }
  }

  if (!(wp->no_surround))
  { 
    monster_decide_action_2(wp);
    if (!(wp->no_hesitate))
    { 
      monster_decide_action_3(wp);
    }
  }

  return 0;
}
