#include <stdio.h>
/* malloc */
#include <stdlib.h>

#include <curses.h>

#include "array.h"
#include "creature.h"
#include "dijkstra.h"

#include "world.h"

world *
world_new(void)
{
  int i;
  world *wp = NULL;

  wp = (world *) malloc(sizeof(world));
  if (wp == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp) failed\n");
    return NULL;
  }

  wp->cr_size = 0;
  wp->cr = NULL;
  wp->size_x = 0;
  wp->size_y = 0;
  wp->grid = NULL;
  wp->player_id = -1;
  wp->scr = NULL;
  wp->scr_size_x = 0;
  wp->scr_size_y = 0;
  wp->should_quit = 0;
  wp->player_decided = 0;
  wp->player_damage = 0;
  wp->dijkp = NULL;
  wp->map_blocked = NULL;
  wp->num_engage = 0;
  wp->engage_id = NULL;
  wp->engage_num_choice = NULL;
  wp->no_surround = 0;
  wp->no_hesitate = 0;

  wp->cr_size = 26;
  wp->cr = (creature **) malloc(sizeof(creature *) * (wp->cr_size));
  if (wp->cr == NULL)
  {
    fprintf(stderr, "world_new: malloc(cr) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->cr_size; i++)
    wp->cr[i] = NULL;

  for (i = 0; i < wp->cr_size; i++)
  {
    wp->cr[i] = creature_new();
    if (wp->cr[i] == NULL)
    {
      fprintf(stderr, "world_new: creature_new(cr[%d]) failed\n", i);
      world_delete(wp);
      wp = NULL;
      return NULL;
    }
    wp->cr[i]->id = i;
  }

  wp->size_x = 79;
  wp->size_y = 23;
  wp->grid = array2_new(wp->size_x, wp->size_y);
  if (wp->grid == NULL)
  {
    fprintf(stderr, "world_new: array2_new(grid) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }

  wp->dijkp = (dijkstra_map **) malloc(sizeof(dijkstra_map *) * 9);
  if (wp->dijkp == NULL)
  {
    fprintf(stderr, "world_new: malloc(dijkp) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < 9; i++)
    wp->dijkp[i] = NULL;

  for (i = 0; i < 9; i++)
  {    
    wp->dijkp[i] = dijkstra_map_new(wp->size_x, wp->size_y);
    if (wp->dijkp[i] == NULL)
    {
      fprintf(stderr, "world_new: dijkstra_map_new(%d) failed\n", i);
      world_delete(wp);
      wp = NULL;
      return NULL;
    }
  }

  wp->map_blocked = array2_new(wp->size_x, wp->size_y);
  if (wp->map_blocked == NULL)
  {
    fprintf(stderr, "world_new: array2_new(map_blocked) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }

  wp->engage_id = (int *) malloc(sizeof(int) * wp->cr_size);
  if (wp->engage_id == NULL)
  {
    fprintf(stderr, "world_new: malloc(engage_id) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }

  wp->engage_num_choice = (int *) malloc(sizeof(int) * wp->cr_size);
  if (wp->engage_num_choice == NULL)
  {
    fprintf(stderr, "world_new: malloc(engage_num_choice) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }

  return wp;
}

void
world_delete(world *wp)
{
  int i;

  if (wp == NULL)
    return;

  if (wp->cr != NULL)
  {
    for (i = 0; i < wp->cr_size; i++)
    {
      if (wp->cr[i] != NULL)
      {
        creature_delete(wp->cr[i]);
        wp->cr[i] = NULL;
      }
    }

    free(wp->cr);
    wp->cr = NULL;
  }
  if (wp->engage_id != NULL)
  {
    free(wp->engage_id);
    wp->engage_id = NULL;
  }
  if (wp->engage_num_choice != NULL)
  {
    free(wp->engage_num_choice);
    wp->engage_num_choice = NULL;
  }

  wp->cr_size = 0;

  if (wp->grid != NULL)
  {
    array2_delete(wp->grid, wp->size_x, wp->size_y);
    wp->grid = NULL;
  }
  if (wp->map_blocked != NULL)
  {
    array2_delete(wp->map_blocked, wp->size_x, wp->size_y);
    wp->map_blocked = NULL;
  }

  wp->size_x = 0;
  wp->size_y = 0;

  wp->player_id = -1;

  if (wp->dijkp != NULL)
  {
    for (i = 0; i < 9; i++)
    {
      if (wp->dijkp[i] == NULL)
        continue;

      dijkstra_map_delete(wp->dijkp[i]);
      wp->dijkp[i] = NULL;
    }

    free(wp->dijkp);
    wp->dijkp = NULL;
  }

  free(wp);
  wp = NULL;
}
